*****************************************************
* DBFOPEN.PRG
* Demonstration of Open_dbf function.
* Caution     : Do not use if the index contains FOR clause.
* Language    : Clipper 5.x
* Author      : Laszlo Tallodi [100324.370@compuserv.com]
* Date        : 10.15.1995
*****************************************************
// See S50009.TXT in CA-CLIPPER Forum library for more details of
// .NTX file structure
// Please send me an email if you find an index key that does not work
// properly with this code.
#include 'error.ch'
#include 'dbstruct.ch'
CLS
IF !Open_Dbf(.T.,"test",{ {"testi1","ctestkey"},;
                          {"testi2","ntestkey"},;
                          {"testi3","ltestkey"},;
                          {"testi4","DTOS(dtestkey)"},;
                          {"testi5","RECNO()"},;
                          {"testi6","deleted()"},;
                          {"testi7","ctestkey+STR(ntestkey,5,3)"},;
                          {"testi8","ntestkey*1000"},;
                          {"testi9","ntestkey*ntestkey2"};
                        };
            )
   //any function
ENDIF
RETURN

********************************************************
* Open_Dbf
* PARAMETERS: lEx    open mode, .T.= exclusive .F. = shared
*             cDbf   dbf  name without extension
*             aIndex array of index files and index keys
* Open a database file and its index files.
* Rebuilds the index file if  the index file is missing,
* corrupted or does not match the database file.
* RETURN    : .T.= succes , .F.=fail
***********************************************************

FUNCTION Open_Dbf(lEx,cDbf,aIndex)
 LOCAL objLocal,nRec,cKeyexpr
 LOCAL oOldhandler:=ERRORBLOCK({|e| Myhandler(e,oOldhandler)})

 // opening
 DO WHILE .T.
    BEGIN SEQUENCE
    IF lEx
       USE &cDbf NEW EXCLUSIVE
    ELSE
       USE &cDbf  NEW
    ENDIF
    RECOVER USING objLocal
      @ 23,0 SAY 'Unable to open '+cDbf+'.dbf'
      ERRORBLOCK(oOldhandler)
      RETURN (.F.)
   END
   EXIT
 ENDDO
 //set indexes
 IF aIndex!=NIL
    nRec:=RECCOUNT()
    FOR ix:=1 TO LEN(aIndex)
        //creating indexfile if the exist
        cKeyExpr:=ALLTRIM(aIndex[ix,2])

        IF !Integrity(aIndex[ix,1],nRec,cKeyExpr,KeySize(cKeyExpr))
           @ 23,0 SAY aIndex[ix,1]+'.ntx is bad or missing. Please, wait ! Rebuilding.'
           DBCREATEIND(aIndex[ix,1],cKeyExpr,{||&(cKeyExpr)})
        ENDIF

    NEXT

    SET INDEX TO
    DO WHILE .T.
    BEGIN SEQUENCE
       FOR ix:=1 TO LEN(aIndex)
              DBSETINDEX(aIndex[ix,1])
       NEXT
    RECOVER  USING objLocal
        @ 23,0 SAY aIndex[ix,1]+'.ntx is bad or missing. Please, wait ! Rebuilding.'
        DBCREATEIND(aIndex[ix,1],cKeyExpr,{||&(cKeyExpr)})
        SET INDEX TO
        LOOP
    END
       exit
    ENDDO
 ENDIF
 SET ORDER TO 1
 ERRORBLOCK(oOldhandler)
RETURN(.T.)
******************************
* MyHandler
* Error handler  of Open_Dbf()
******************************
STATIC FUNCTION MyHandler(obj,handler)

 IF obj:severety!=ES_CATASTROPHIC
    IF (obj:gencode==EG_OPEN .OR. obj:gencode==EG_CORRUPTION)
         BREAK obj
    ENDIF
 ENDIF
RETURN (EVAL(handler,obj))
**************************
FUNCTION KeySize(cKeyExpr)
 LOCAL xValue,aKeySize,aStruct
 LOCAL nElem,nFieldSize,nWpoint

 aStruct:=DBSTRUCT()
 IF (nElem:=ASCAN(aStruct,{|aVal|aVal[DBS_NAME]==UPPER(cKeyExpr)}))!=0

 //if the key expression is a field

     nFieldSize:=aStruct[nElem,DBS_LEN]
     nFieldDec :=aStruct[nElem,DBS_DEC]
     aKeySize:={nFieldSize,nFieldDec}

 ELSE
    //Evaluating the key expression
    xValue:=&cKeyExpr

    DO CASE
       CASE  VALTYPE(xValue)=='N'
             xValue:=STR(xValue)
             nWpoint:=AT('.',xValue)
             aKeySize:={LEN(xValue),IF(nWpoint==0,0,LEN(xValue)-nWpoint)}
       CASE  VALTYPE(xValue)=='L'
             aKeySize:={1,0}
       CASE  VALTYPE(xValue)=='D'
             aKeySize:={8,0}
       CASE  VALTYPE(xValue)=='C'
            aKeySize:={LEN(xValue),0}
       OTHERWISE
       //?
    ENDCASE

 ENDIF

RETURN (aKeySize)
********************************************************
* Integrity
* PARAMETERS: cIndf     index file name
*                       (default extension .NTX.)
*             nRec      number of records in  DBF
*             cKey      index key expression
*             aKeySize  array of key size
*                       first element :   key length
*                       second element:   number of decimals in key
* RETURN   :  .T.:  NTX matchs DBF, .F.: other
********************************************************

STATIC FUNCTION Integrity(cIndf,nRec,cKey,aKeySize)
 LOCAL cFname :=cIndf+'.ntx'
 LOCAL nHandle,lRetv:=.F.

 IF (nHandle := Fopen(cFname))!=-1

    IF  okItemNumber(nHandle,nRec) .and. ;
        okKeyExpr(nHandle,cKey) .and.;
        okKeySize(nHandle,aKeySize)

            lRetv:=.T.

    ENDIF

     Fclose(nHandle)

 ENDIF

 RETURN (lRetv)

******************************************
STATIC FUNCTION okItemNumber(nHandle,nRec)
 LOCAL cBuffW:= space(2)
 LOCAL nIRec := 0

 // scanning  the index file for
 // counting the items
 // the first word(two bytes) of the subsequent index page contains
 // the number of used entries on the index page

 Fseek(nHandle,1024,0)
 Fread(nHandle,@cBuffW,2)
 nIRec:=BIN2W(cBuffW)

 DO WHILE FERROR(nHandle)==0

    Fseek(nHandle,1022,1)

    IF Fread(nHandle,@cBuffW,2)!=2

       EXIT

    ENDIF

    nIRec+=BIN2W(cBuffW)

 ENDDO

RETURN (nRec==nIRec)
***************************************
STATIC FUNCTION okKeyExpr(nHandle,cKey)
 LOCAL cBuffB := space(1)
 LOCAL cIKey:=SPACE(0)

 // reading the key expression beginning at offset 16h
 // and followed by CHR(0) byte

 Fseek(nHandle,22,0)
 Fread(nHandle,@cBuffB,1)

 DO WHILE FERROR(nHandle)==0 .and. ASC(cBuffB)!=0

    cIKey+=cBuffB

    IF Fread(nHandle,@cBuffB,1)!=1
       EXIT
    ENDIF

 ENDDO

RETURN (UPPER(cKey)==UPPER(cIkey))
*************************
STATIC FUNCTION okKeySize(nHandle,aKeySize)
 LOCAL cBuffW := space(2)
 LOCAL cIKey:=SPACE(0)
 LOCAL nIKeySize:=0,nIKeyDec:=0

 //Reading the key size  at offset 0Eh

 Fseek(nHandle,14,0)
 Fread(nHandle,@cbuffW,2)
 nIKeySize:=BIN2W(cBuffW)

 //Reading number of decimals in key at offset 10h

 Fseek(nHandle,16,0)
 Fread(nHandle,@cbuffW,2)
 nIKeyDec:=BIN2W(cBuffW)

RETURN (aKeySize[1]==nIKeysize .and. aKeySize[2]==nIKeyDec)
